// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/16AIO168/16aio168_Linux_2.x.x.x_DN/driver/device.c $
// $Rev: 53667 $
// $Date: 2023-08-31 16:09:35 -0500 (Thu, 31 Aug 2023) $

// 16AIO168: Device Driver: source file

#include "main.h"



// variables ******************************************************************

const gsc_dev_id_t	dev_id_list[]	=
{
	// model	Vendor	Device	SubVen	SubDev	type

	{ "16AIO168",	0x10B5, 0x9080, 0x10B5, 0x2879,	GSC_DEV_TYPE_16AIO168	},
	{ NULL }
};



//*****************************************************************************
static void _fw_rev_data_init(dev_data_t* dev)
{
	u32	frr;

	frr					= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_frr_32);
	dev->cache.fw_rev	= frr & 0xFFFF;
}



/******************************************************************************
*
*	Function:	dev_device_create
*
*	Purpose:
*
*		Do everything needed to setup and use the given device.
*
*	Arguments:
*
*		dev		The structure to initialize.
*
*	Returned:
*
*		0		All is well.
*		< 0		An appropriate error status.
*
******************************************************************************/

int dev_device_create(dev_data_t* dev)
{
	static const gsc_bar_maps_t	bar_map	=
	{
		{
			// mem	io	rw
			{ 1,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 0: PLX registers, memory mapped
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 1: PLX registers, I/O mapped
			{ 1,	0,	GSC_REG_TYPE_ACCESS_RW },	// BAR 2: GSC registers, memory mapped
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 3: unused
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 4: unused
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO }	// BAR 5: unused
		}
	};

	u32	dma;
	int	ret;

	for (;;)	// A convenience loop.
	{
		// Verify some macro contents.
		ret	= gsc_macro_test_base_name(AIO168_BASE_NAME);
		if (ret)	break;

		ret	= gsc_macro_test_model();
		if (ret)	break;

		// PCI setup.
		ret	= os_pci_dev_enable(&dev->pci);
		if (ret)	break;

		ret	= os_pci_master_set(&dev->pci);
		if (ret)	break;

		// Control ISR access to the device and data structure.
		ret	= os_spinlock_create(&dev->spinlock);
		if (ret)	break;

		// Control access to the device and data structure.
		ret	= os_sem_create(&dev->sem);
		if (ret)	break;

		// Access the BAR regions.
		ret	= gsc_bar_create(dev, &dev->bar, &bar_map);
		if (ret)	break;

		// Firmware access.
		dev->vaddr.gsc_bcr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AIO168_GSC_BCR));
		dev->vaddr.gsc_docr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AIO168_GSC_DOCR));
		dev->vaddr.gsc_frr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AIO168_GSC_FRR));
		dev->vaddr.gsc_ibcr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AIO168_GSC_IBCR));
		dev->vaddr.gsc_icr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AIO168_GSC_ICR));
		dev->vaddr.gsc_idbr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AIO168_GSC_IDBR));
		dev->vaddr.gsc_ragr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AIO168_GSC_RAGR));
		dev->vaddr.gsc_rbgr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AIO168_GSC_RBGR));
		dev->vaddr.gsc_obcr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AIO168_GSC_OBCR));
		dev->vaddr.gsc_odbr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AIO168_GSC_ODBR));
		dev->vaddr.gsc_sscr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AIO168_GSC_SSCR));

		// Data cache initialization.
		dev->cache.autocal_ms		= 3000;
		dev->cache.chan_ai_max		= 16;
		dev->cache.chan_ao_max		= 8;
		dev->cache.fgen_max			= 300000;
		dev->cache.fgen_min			= 1;
		dev->cache.fifo_size_rx		= _32K;
		dev->cache.fifo_size_tx		= _32K;
		dev->cache.fref_default		= 30000000L;
		dev->cache.fsamp_max		= 300000;
		dev->cache.fsamp_min		= 1;
		dev->cache.initialize_ms	= 3;
		dev->cache.nrate_max		= 0xFFFF;
		dev->cache.nrate_min		= 100;

		_fw_rev_data_init(dev);

		// Initialize additional resources.
		ret	= dev_irq_create(dev);
		if (ret)	break;

		ret	= dev_io_create(dev);
		if (ret)	break;

		dma	= GSC_DMA_SEL_STATIC
			| GSC_DMA_CAP_BMDMA_READ
			| GSC_DMA_CAP_BMDMA_WRITE;
		ret	= gsc_dma_create(dev, dma, dma);

		break;
	}

	return(ret);
}



/******************************************************************************
*
*	Function:	dev_device_destroy
*
*	Purpose:
*
*		Do everything needed to release the referenced device.
*
*	Arguments:
*
*		dev		The partial data for the device of interest.
*
*	Returned:
*
*		None.
*
******************************************************************************/

void dev_device_destroy(dev_data_t* dev)
{
	if (dev)
	{
		gsc_dma_destroy(dev);
		dev_io_destroy(dev);
		dev_irq_destroy(dev);
		gsc_bar_destroy(&dev->bar);
		os_sem_destroy(&dev->sem);
		os_spinlock_destroy(&dev->spinlock);
		os_pci_master_clear(&dev->pci);
		os_pci_dev_disable(&dev->pci);
	}
}


