// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/16AIO168/16aio168_Linux_2.x.x.x_DN/driver/ioctl.c $
// $Rev: 53667 $
// $Date: 2023-08-31 16:09:35 -0500 (Thu, 31 Aug 2023) $

// 16AIO168: Device Driver: source file

#include "main.h"



// macros *********************************************************************

#define	AUTOCAL_START			D13		// BCR
#define	AUTOCAL_PASS			D14		// BCR
#define	INIT_START				D15		// BCR

#define	IRQ0_MASK				0x0007	// ICR
#define	IRQ0_INIT_DONE			0x0000	// ICR
#define	IRQ0_AUTOCAL_DONE		0x0001	// ICR
#define	IRQ0_ACTIVE				0x0008	// ICR
#define	IRQ1_ACTIVE				0x0080	// ICR
#define	IRQ2_ACTIVE				0x0800	// ICR

#define	SSCR_SCAN_1				D11
#define	SSCR_SCAN_2				D17
#define	SSCR_SCAN_MULTI			(D0 | D1)



//*****************************************************************************
static int _query(dev_data_t* dev, s32* arg)
{
	int		ret	= 0;

	switch (arg[0])
	{
		default:	arg[0]	= AIO168_IOCTL_QUERY_ERROR;
					ret		= -EINVAL;
					break;

		case AIO168_QUERY_AUTOCAL_MS:	arg[0]	= dev->cache.autocal_ms;	break;
		case AIO168_QUERY_CHAN_AI_MAX:	arg[0]	= dev->cache.chan_ai_max;	break;
		case AIO168_QUERY_CHAN_AO_MAX:	arg[0]	= dev->cache.chan_ao_max;	break;
		case AIO168_QUERY_COUNT:		arg[0]	= AIO168_IOCTL_QUERY_LAST;	break;
		case AIO168_QUERY_DEVICE_TYPE:	arg[0]	= dev->board_type;			break;
		case AIO168_QUERY_FIFO_SIZE_RX:	arg[0]	= dev->cache.fifo_size_rx;	break;
		case AIO168_QUERY_FIFO_SIZE_TX:	arg[0]	= dev->cache.fifo_size_tx;	break;
		case AIO168_QUERY_FGEN_MAX:		arg[0]	= dev->cache.fgen_max;		break;
		case AIO168_QUERY_FGEN_MIN:		arg[0]	= dev->cache.fgen_min;		break;
		case AIO168_QUERY_FREF_DEFAULT:	arg[0]	= dev->cache.fref_default;	break;
		case AIO168_QUERY_FSAMP_MAX:	arg[0]	= dev->cache.fsamp_max;		break;
		case AIO168_QUERY_FSAMP_MIN:	arg[0]	= dev->cache.fsamp_min;		break;
		case AIO168_QUERY_INIT_MS:		arg[0]	= dev->cache.initialize_ms;	break;
		case AIO168_QUERY_NRATE_MAX:	arg[0]	= dev->cache.nrate_max;		break;
		case AIO168_QUERY_NRATE_MIN:	arg[0]	= dev->cache.nrate_min;		break;
	}

	return(ret);
}



//*****************************************************************************
static int _init_start(dev_data_t* dev, void* arg)
{
	// Initiate initialization.
	os_reg_mem_tx_u32(NULL, dev->vaddr.gsc_bcr_32, INIT_START);
	return(0);
}



//*****************************************************************************
int initialize_ioctl(dev_data_t* dev, void* arg)
{
	int			i;
	int			mask;
	long		ms			= dev->cache.initialize_ms + 5000;
	long		ms_total	= ms;
	u32			reg;
	int			ret			= 0;
	os_sem_t	sem;
	int			tmp;
	VADDR_T		va			= dev->vaddr.gsc_icr_32;
	int			value;
	gsc_wait_t	wt;

	if ((dev->irq.opened) && (gsc_global.driver_unloading == 0))
	{
		ms_total	*= 2;

		// Safely select the Initialize Done interrupt.
		mask	= IRQ0_MASK | IRQ0_ACTIVE | IRQ1_ACTIVE | IRQ2_ACTIVE;
		value	= IRQ0_INIT_DONE;

		if (dev->cache.fw_rev > 0x903)
		{
			// In newer firmware these don't cause interrupts.
			value	|= IRQ1_ACTIVE | IRQ2_ACTIVE;
		}

		os_reg_mem_mx_u32(dev, va, value, mask);

		// Wait for the local interrupt.
		os_sem_create(&sem);	// dummy, required for wait operations.
		memset(&wt, 0, sizeof(wt));
		wt.flags		= GSC_WAIT_FLAG_INTERNAL;
		wt.gsc			= AIO168_WAIT_GSC_IDLE_INIT;
		wt.timeout_ms	= ms;
		ret				= gsc_wait_event(dev, &wt, _init_start, NULL, &sem);
		os_sem_destroy(&sem);

		if (wt.flags & GSC_WAIT_FLAG_TIMEOUT)
		{
			ret	= ret ? ret : -ETIMEDOUT;
			printf(	"%s: INITILIZE DONE IRQ TIMED OUT AFTER %ld ms.\n",
					dev->model,
					(long) ms);
		}
	}
	else
	{
		_init_start(dev, NULL);
	}

	va	= dev->vaddr.gsc_bcr_32;

	// Manually wait for completion as the IRQ wait may have ended early.
	tmp	= gsc_poll_u32(dev, ms, va, INIT_START, 0);

	if (tmp)
	{
		ret	= ret ? ret : tmp;
		printf(	"%s: INITIALIZATION DID NOT COMPLETE WITHIN %ld ms.\n",
				dev->model,
				ms_total);
	}

	// Manually check for completion as the IRQ wait may have ended early.
	reg	= os_reg_mem_rx_u32(dev, va);

	if (reg & INIT_START)
	{
		ret	= ret ? ret : -ETIMEDOUT;
		printf(	"%s: INITILIZE STILL ACTIVE AFTER %ld ms.\n",
				dev->model,
				ms_total);
	}

	// Initialize the software settings.

	for (i = 0; i < DEV_IO_STREAM_QTY; i++)
	{
		if (dev->io.io_streams[i])
		{
			if (dev->io.io_streams[i]->dev_io_sw_init)
			{
				(dev->io.io_streams[i]->dev_io_sw_init)(dev, dev->io.io_streams[i]);
			}
		}
	}

	return(ret);
}



//*****************************************************************************
static int _autocal_start(dev_data_t* dev, void* arg)
{
	// Initiate autocalibration.
	os_reg_mem_mx_u32(NULL, dev->vaddr.gsc_bcr_32, AUTOCAL_START, AUTOCAL_START);
	return(0);
}



//*****************************************************************************
static int _autocal(dev_data_t* dev, void* arg)
{
	u32			mask;
	long		ms		= dev->cache.autocal_ms + 5000;
	u32			reg;
	int			ret;
	os_sem_t	sem;
	int			tmp;
	VADDR_T		va		= dev->vaddr.gsc_bcr_32;
	u32			value;
	gsc_wait_t	wt;

	// Safely select the Autocal Done interrupt.
	mask	= IRQ0_ACTIVE | IRQ1_ACTIVE | IRQ2_ACTIVE | IRQ0_MASK;
	value	= IRQ0_AUTOCAL_DONE;

	if (dev->cache.fw_rev > 0x903)
	{
		// In newer firmware these don't cause interrupts.
		value	|= IRQ1_ACTIVE | IRQ2_ACTIVE;
	}

	irq_icr_mod(dev, value, mask);

	// Wait for the local interrupt.
	os_sem_create(&sem);	// dummy, required for wait operations.
	memset(&wt, 0, sizeof(wt));
	wt.flags		= GSC_WAIT_FLAG_INTERNAL;
	wt.gsc			= AIO168_WAIT_GSC_AUTOCAL_DONE;
	wt.timeout_ms	= ms;
	ret				= gsc_wait_event(dev, &wt, _autocal_start, NULL, &sem);
	os_sem_destroy(&sem);

	if (wt.flags & GSC_WAIT_FLAG_TIMEOUT)
	{
		ret	= ret ? ret : -ETIMEDOUT;
		printf(	"%s: AUTOCALIBRATE DONE IRQ TIMED OUT AFTER %ld ms.\n",
				dev->model,
				ms);
	}

	// Manually wait for completion in case something terminates our wait early.
	tmp	= gsc_poll_u32(dev, ms, va, AUTOCAL_START, 0);

	if (tmp)
	{
		ms	*= 2;
		ret	= ret ? ret : tmp;
		printf(	"%s: AUTOCALIBRATION DID NOT COMPLETE WITHIN %ld ms.\n",
				dev->model,
				ms);
	}

	// Manually check for completion as the IRQ wait may have ended early.
	reg	= os_reg_mem_rx_u32(dev, va);

	if (reg & AUTOCAL_START)
	{
		ret	= ret ? ret : -ETIMEDOUT;
		printf(	"%s: AUTOCALIBRATION STILL ACTIVE AFTER %ld ms.\n",
				dev->model,
				ms);
	}

	// Final results.
	reg	= os_reg_mem_rx_u32(dev, va);

	if ((reg & AUTOCAL_PASS) == 0)
	{
		ret	= ret ? ret : -EIO;
		printf(	"%s: AUTOCALIBRATION FAILED (%ld ms).\n",
				dev->model,
				ms);
	}

	return(ret);
}



//*****************************************************************************
static int _autocal_status(dev_data_t* dev, s32* arg)
{
	u32	reg;

	reg	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bcr_32);

	if (reg & AUTOCAL_START)
		arg[0]	= AIO168_AUTOCAL_STATUS_ACTIVE;
	else if (reg & AUTOCAL_PASS)
		arg[0]	= AIO168_AUTOCAL_STATUS_PASS;
	else
		arg[0]	= AIO168_AUTOCAL_STATUS_FAIL;

	return(0);
}



//*****************************************************************************
static int _ai_buf_clear(dev_data_t* dev, void* arg)
{
	#define	CLEAR	D15		// IBCR

	int		i;
	u32		reg;
	int		ret	= 0;
	VADDR_T	va	= dev->vaddr.gsc_ibcr_32;

	// Clear the buffer.
	os_reg_mem_mx_u32(dev, va, CLEAR, CLEAR);

	// Wait for the bit to clear.

	for (i = 0;; i++)
	{
		reg	= os_reg_mem_rx_u32(dev, va);

		if ((reg & CLEAR) == 0)
			break;

		if (i >= 250)
		{
			ret	= -EINVAL;
			printf(	"%s: The analog input buffer took too long to clear.\n",
					DEV_NAME);
			break;
		}
	}

	return(ret);
}



//*****************************************************************************
static int _ai_buf_thr_lvl(dev_data_t* dev, void* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(dev, arg, 0, 0x7FFF, dev->vaddr.gsc_ibcr_32, 14, 0);
	return(ret);
}



//*****************************************************************************
static int _ai_buf_thr_sts(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AIO168_AI_BUF_THR_STS_CLEAR,
		AIO168_AI_BUF_THR_STS_SET,
		-1	// terminate list
	};

	int	ret;

	arg[0]	= -1;
	ret		= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_ibcr_32, 16, 16);
	return(ret);
}



//*****************************************************************************
static int _ai_mode(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AIO168_AI_MODE_DIFF,
		AIO168_AI_MODE_SINGLE,
		AIO168_AI_MODE_ZERO,
		AIO168_AI_MODE_VREF,
		AIO168_AI_MODE_AO_0,
		AIO168_AI_MODE_AO_1,
		AIO168_AI_MODE_AO_2,
		AIO168_AI_MODE_AO_3,
		AIO168_AI_MODE_AO_4,
		AIO168_AI_MODE_AO_5,
		AIO168_AI_MODE_AO_6,
		AIO168_AI_MODE_AO_7,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 3, 0);
	return(ret);
}



//*****************************************************************************
static int _ai_scan_1_chan(dev_data_t* dev, void* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(dev, arg, 0, 15, dev->vaddr.gsc_sscr_32, 16, 12);
	return(ret);
}



//*****************************************************************************
static int _ai_scan_clk_src(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AIO168_AI_SCAN_CLK_SRC_RAG,
		AIO168_AI_SCAN_CLK_SRC_RBG,
		AIO168_AI_SCAN_CLK_SRC_EXT,
		AIO168_AI_SCAN_CLK_SRC_BCR,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_sscr_32, 3, 2);
	return(ret);
}



//*****************************************************************************
static int _ai_scan_size(dev_data_t* dev, void* arg)
{
	s32*	ptr		= (s32*) arg;
	int		ret		= 0;
	u32		sscr	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_sscr_32);

	if (ptr[0] == -1)
	{
		if (sscr & SSCR_SCAN_2)
		{
			ptr[0]	= AIO168_AI_SCAN_SIZE_0_1;
		}
		else if (sscr & SSCR_SCAN_1)
		{
			ptr[0]	= AIO168_AI_SCAN_SIZE_SINGLE;
		}
		else
		{
			switch (sscr & 0x3)
			{
				default:	ret		= -EIO;						break;
				case 0x0:	ptr[0]	= AIO168_AI_SCAN_SIZE_0_3;	break;
				case 0x1:	ptr[0]	= AIO168_AI_SCAN_SIZE_0_7;	break;
				case 0x2:	ptr[0]	= AIO168_AI_SCAN_SIZE_0_15;	break;	// Single Ended mode only.
			}
		}
	}
	else
	{
		switch (ptr[0])
		{
			default:

				ret	= -EIO;
				break;

			case AIO168_AI_SCAN_SIZE_SINGLE:

				os_reg_mem_mx_u32(dev, dev->vaddr.gsc_sscr_32, SSCR_SCAN_1, SSCR_SCAN_1);
				os_reg_mem_mx_u32(dev, dev->vaddr.gsc_sscr_32, 0, SSCR_SCAN_2);
				break;

			case AIO168_AI_SCAN_SIZE_0_1:

				os_reg_mem_mx_u32(dev, dev->vaddr.gsc_sscr_32, 0, SSCR_SCAN_1);
				os_reg_mem_mx_u32(dev, dev->vaddr.gsc_sscr_32, SSCR_SCAN_2, SSCR_SCAN_2);
				break;

			case AIO168_AI_SCAN_SIZE_0_3:

				os_reg_mem_mx_u32(dev, dev->vaddr.gsc_sscr_32, 0, SSCR_SCAN_1);
				os_reg_mem_mx_u32(dev, dev->vaddr.gsc_sscr_32, 0, SSCR_SCAN_2);
				os_reg_mem_mx_u32(dev, dev->vaddr.gsc_sscr_32, 0, SSCR_SCAN_MULTI);
				break;

			case AIO168_AI_SCAN_SIZE_0_7:

				os_reg_mem_mx_u32(dev, dev->vaddr.gsc_sscr_32, 0, SSCR_SCAN_1);
				os_reg_mem_mx_u32(dev, dev->vaddr.gsc_sscr_32, 0, SSCR_SCAN_2);
				os_reg_mem_mx_u32(dev, dev->vaddr.gsc_sscr_32, 1, SSCR_SCAN_MULTI);
				break;

			case AIO168_AI_SCAN_SIZE_0_15:	// Single Ended mode only.

				os_reg_mem_mx_u32(dev, dev->vaddr.gsc_sscr_32, 0, SSCR_SCAN_1);
				os_reg_mem_mx_u32(dev, dev->vaddr.gsc_sscr_32, 0, SSCR_SCAN_2);
				os_reg_mem_mx_u32(dev, dev->vaddr.gsc_sscr_32, 2, SSCR_SCAN_MULTI);
				break;
		}
	}

	return(ret);
}



//*****************************************************************************
static int _ai_sync(dev_data_t* dev, void* arg)
{
	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_bcr_32, D12, D12);
	return(0);
}



//*****************************************************************************
static int _ao_buf_clear(dev_data_t* dev, void* arg)
{
	#define	CLEAR	D15		// OBCR

	int		i;
	u32		reg;
	int		ret	= 0;
	VADDR_T	va	= dev->vaddr.gsc_obcr_32;

	// Clear the buffer.
	os_reg_mem_mx_u32(dev, va, CLEAR, CLEAR);

	// Wait for the bit to clear.

	for (i = 0;; i++)
	{
		reg	= os_reg_mem_rx_u32(dev, va);

		if ((reg & CLEAR) == 0)
			break;

		if (i >= 250)
		{
			ret	= -EINVAL;
			printf(	"%s: The analog output buffer took too long to clear.\n",
					DEV_NAME);
			break;
		}
	}

	return(ret);
}



//*****************************************************************************
static int _ao_buf_thr_lvl(dev_data_t* dev, void* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(dev, arg, 0, 0x7FFF, dev->vaddr.gsc_obcr_32, 14, 0);
	return(ret);
}



//*****************************************************************************
static int _ao_buf_thr_sts(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AIO168_AO_BUF_THR_STS_CLEAR,
		AIO168_AO_BUF_THR_STS_SET,
		-1	// terminate list
	};

	int	ret;

	arg[0]	= -1;
	ret		= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_obcr_32, 16, 16);
	return(ret);
}



//*****************************************************************************
static int _ao_burst_clk_src(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AIO168_AO_BURST_CLK_SRC_RAG,
		AIO168_AO_BURST_CLK_SRC_RBG,
		AIO168_AO_BURST_CLK_SRC_EXT,
		AIO168_AO_BURST_CLK_SRC_BCR,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_sscr_32, 7, 6);
	return(ret);
}



//*****************************************************************************
static int _ao_burst_enable(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AIO168_AO_BURST_ENABLE_NO,
		AIO168_AO_BURST_ENABLE_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 9, 9);
	return(ret);
}



//*****************************************************************************
static int _ao_clk_src(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AIO168_AO_CLK_SRC_RAG,
		AIO168_AO_CLK_SRC_RBG,
		AIO168_AO_CLK_SRC_EXT,
		AIO168_AO_CLK_SRC_DISABLE,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_sscr_32, 5, 4);
	return(ret);
}



//*****************************************************************************
static int _ao_looping(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AIO168_AO_LOOPING_DISABLE,
		AIO168_AO_LOOPING_ENABLE,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 10, 10);
	return(ret);
}



//*****************************************************************************
static int _ao_sync(dev_data_t* dev, void* arg)
{
	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_bcr_32, D11, D11);
	return(0);
}



//*****************************************************************************
static int _ao_timing(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AIO168_AO_TIMING_SEQ,
		AIO168_AO_TIMING_SIMUL,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 8, 8);
	return(ret);
}



//*****************************************************************************
static int _data_format(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AIO168_DATA_FORMAT_2S_COMP,
		AIO168_DATA_FORMAT_OFF_BIN,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 6, 6);
	return(ret);
}



//*****************************************************************************
static int _dio_write(dev_data_t* dev, void* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(dev, arg, 0, 0xF, dev->vaddr.gsc_docr_32, 3, 0);
	return(ret);
}



//*****************************************************************************
static int _ext_sync_src(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AIO168_AI_EXT_SYNC_SRC_AISC,
		AIO168_AI_EXT_SYNC_SRC_AOS,
		AIO168_AI_EXT_SYNC_SRC_EXT,
		AIO168_AI_EXT_SYNC_SRC_DISABLE,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_sscr_32, 9, 8);
	return(ret);
}



//*****************************************************************************
static int _irq0_sel(dev_data_t* dev, s32* arg)
{
	u32	reg;
	int	ret	= 0;
	u32	val;

	if (arg[0] == -1)
	{
		// Retrieve the current setting.
		reg		= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_icr_32);
		arg[0]	= GSC_FIELD_DECODE(reg, 2, 0);
	}
	else
	{
		// Validate the option value passed in.

		switch (arg[0])
		{
			default:

				ret	= -EINVAL;
				break;

			case AIO168_IRQ0_IDLE_INIT:
			case AIO168_IRQ0_AUTOCAL_DONE:

				break;
		}

		if (ret == 0)
		{
			// Clear the status bit and apply the new setting.
			// Do not clear the IRQ1 or IRQ2 status bit.

			if (dev->cache.fw_rev <= 0x903)
				val	= arg[0];
			else
				val	= arg[0] | 0x880;

			os_reg_mem_mx_u32(dev, dev->vaddr.gsc_icr_32, val, 0x88F);
		}
	}

	return(ret);
}



//*****************************************************************************
static int _irq1_sel(dev_data_t* dev, s32* arg)
{
	u32	reg;
	int	ret	= 0;
	u32	val;

	if (arg[0] == -1)
	{
		// Retrieve the current setting.
		reg		= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_icr_32);
		arg[0]	= GSC_FIELD_DECODE(reg, 6, 4);
	}
	else
	{
		// Validate the option value passed in.

		switch (arg[0])
		{
			default:

				ret	= -EINVAL;
				break;

			case AIO168_IRQ1_IDLE:
			case AIO168_IRQ1_AI_BUF_THR_L2H:
			case AIO168_IRQ1_AI_BUF_THR_H2L:

				break;
		}

		if (ret == 0)
		{
			// Clear the status bit and apply the new setting.
			// Do not clear the IRQ0 or IRQ2 status bit.

			if (dev->cache.fw_rev <= 0x903)
				val	= (arg[0] << 4);
			else
				val	= (arg[0] << 4) | 0x808;

			os_reg_mem_mx_u32(dev, dev->vaddr.gsc_icr_32, val, 0x8F8);
		}
	}

	return(ret);
}



//*****************************************************************************
static int _irq2_sel(dev_data_t* dev, s32* arg)
{
	u32	reg;
	int	ret	= 0;
	u32	val;

	if (arg[0] == -1)
	{
		// Retrieve the current setting.
		reg		= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_icr_32);
		arg[0]	= GSC_FIELD_DECODE(reg, 10, 8);
	}
	else
	{
		// Validate the option value passed in.

		switch (arg[0])
		{
			default:

				ret	= -EINVAL;
				break;

			case AIO168_IRQ2_IDLE:
			case AIO168_IRQ2_AO_BUF_THR_L2H:
			case AIO168_IRQ2_AO_BUF_THR_H2L:
			case AIO168_IRQ2_AO_BURST_DONE:

				break;
		}

		if (ret == 0)
		{
			// Clear the status bit and apply the new setting.
			// Do not clear the IRQ0 or IRQ1 status bit.

			if (dev->cache.fw_rev <= 0x903)
				val	= (arg[0] << 8);
			else
				val	= (arg[0] << 8) | 0x088;

			os_reg_mem_mx_u32(dev, dev->vaddr.gsc_icr_32, val, 0xF88);
		}
	}

	return(ret);
}



//*****************************************************************************
static int _rag_enable(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AIO168_GEN_ENABLE_NO,
		AIO168_GEN_ENABLE_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_ragr_32, 16, 16);
	return(ret);
}



//*****************************************************************************
static int _rag_nrate(dev_data_t* dev, void* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(
			dev,
			arg,
			dev->cache.nrate_min,
			dev->cache.nrate_max,
			dev->vaddr.gsc_ragr_32,
			15,
			0);

	return(ret);
}



//*****************************************************************************
static int _range(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AIO168_RANGE_2_5V,
		AIO168_RANGE_5V,
		AIO168_RANGE_10V,
		-1	// terminate list
	};

	u32		bcr;
	s32*	ptr	= (s32*) arg;
	int		ret	= 0;

	if (ptr[0] == -1)
	{
		bcr	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bcr_32);

		switch (bcr & 0x30)
		{
			default:	ret		= -EINVAL;				break;
			case 0x00:	ptr[0]	= AIO168_RANGE_2_5V;	break;
			case 0x10:	ptr[0]	= AIO168_RANGE_5V;		break;
			case 0x20:
			case 0x30:	ptr[0]	= AIO168_RANGE_10V;		break;
		}
	}
	else
	{
		ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 5, 4);
	}

	return(ret);
}



//*****************************************************************************
static int _rbg_clk_src(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AIO168_RBG_CLK_SRC_MASTER,
		AIO168_RBG_CLK_SRC_RAG,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_sscr_32, 10, 10);
	return(ret);
}



//*****************************************************************************
static int _rbg_enable(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AIO168_GEN_ENABLE_NO,
		AIO168_GEN_ENABLE_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_rbgr_32, 16, 16);
	return(ret);
}



//*****************************************************************************
static int _rbg_nrate(dev_data_t* dev, void* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(
			dev,
			arg,
			dev->cache.nrate_min,
			dev->cache.nrate_max,
			dev->vaddr.gsc_rbgr_32,
			15,
			0);
	return(ret);
}



//*****************************************************************************
static int _rx_io_abort(dev_data_t* dev, s32* arg)
{
	arg[0]	= gsc_read_abort_active_xfer(dev, &dev->io.rx);
	return(0);
}



//*****************************************************************************
static int _rx_io_mode(dev_data_t* dev, s32* arg)
{
	static const s32	list[]	=
	{
		GSC_IO_MODE_PIO,
		GSC_IO_MODE_BMDMA,
		-1
	};

	int	ret;

	ret	= gsc_s32_list_var(arg, list, &dev->io.rx.io_mode);
	return(ret);
}



//*****************************************************************************
static int _rx_io_timeout(dev_data_t* dev, s32* arg)
{
	int	ret;

	ret	= gsc_s32_range_var(
			arg,
			AIO168_IO_TIMEOUT_MIN,
			AIO168_IO_TIMEOUT_INFINITE,
			&dev->io.rx.timeout_s);
	return(ret);
}



//*****************************************************************************
static int _tx_io_abort(dev_data_t* dev, s32* arg)
{
	arg[0]	= gsc_write_abort_active_xfer(dev, &dev->io.tx);
	return(0);
}



//*****************************************************************************
static int _tx_io_mode(dev_data_t* dev, s32* arg)
{
	static const s32	list[]	=
	{
		GSC_IO_MODE_PIO,
		GSC_IO_MODE_BMDMA,
		-1
	};

	int	ret;

	ret	= gsc_s32_list_var(arg, list, &dev->io.tx.io_mode);
	return(ret);
}



//*****************************************************************************
static int _tx_io_timeout(dev_data_t* dev, s32* arg)
{
	int	ret;

	ret	= gsc_s32_range_var(
			arg,
			AIO168_IO_TIMEOUT_MIN,
			AIO168_IO_TIMEOUT_INFINITE,
			&dev->io.tx.timeout_s);
	return(ret);
}



// variables ******************************************************************

const gsc_ioctl_t	dev_ioctl_list[]	=
{
	{ AIO168_IOCTL_REG_READ,			(void*) gsc_reg_read_ioctl		},
	{ AIO168_IOCTL_REG_WRITE,			(void*) gsc_reg_write_ioctl		},
	{ AIO168_IOCTL_REG_MOD,				(void*) gsc_reg_mod_ioctl		},
	{ AIO168_IOCTL_QUERY,				(void*) _query					},
	{ AIO168_IOCTL_INITIALIZE,			(void*) initialize_ioctl		},
	{ AIO168_IOCTL_AUTOCAL,				(void*) _autocal				},
	{ AIO168_IOCTL_AUTOCAL_STATUS,		(void*) _autocal_status			},
	{ AIO168_IOCTL_AI_BUF_CLEAR,		(void*) _ai_buf_clear			},
	{ AIO168_IOCTL_AI_BUF_THR_LVL,		(void*) _ai_buf_thr_lvl			},
	{ AIO168_IOCTL_AI_BUF_THR_STS,		(void*) _ai_buf_thr_sts			},
	{ AIO168_IOCTL_AI_MODE,				(void*) _ai_mode				},
	{ AIO168_IOCTL_AI_SCAN_1_CHAN,		(void*) _ai_scan_1_chan			},
	{ AIO168_IOCTL_AI_SCAN_CLK_SRC,		(void*) _ai_scan_clk_src		},
	{ AIO168_IOCTL_AI_SCAN_SIZE,		(void*) _ai_scan_size			},
	{ AIO168_IOCTL_AI_SYNC,				(void*) _ai_sync				},
	{ AIO168_IOCTL_AO_BUF_CLEAR,		(void*) _ao_buf_clear			},
	{ AIO168_IOCTL_AO_BUF_THR_LVL,		(void*) _ao_buf_thr_lvl			},
	{ AIO168_IOCTL_AO_BUF_THR_STS,		(void*) _ao_buf_thr_sts			},
	{ AIO168_IOCTL_AO_BURST_CLK_SRC,	(void*) _ao_burst_clk_src		},
	{ AIO168_IOCTL_AO_BURST_ENABLE,		(void*) _ao_burst_enable		},
	{ AIO168_IOCTL_AO_CLK_SRC,			(void*) _ao_clk_src				},
	{ AIO168_IOCTL_AO_LOOPING,			(void*) _ao_looping				},
	{ AIO168_IOCTL_AO_SYNC,				(void*) _ao_sync				},
	{ AIO168_IOCTL_AO_TIMING,			(void*) _ao_timing				},
	{ AIO168_IOCTL_DATA_FORMAT,			(void*) _data_format			},
	{ AIO168_IOCTL_DIO_WRITE,			(void*) _dio_write				},
	{ AIO168_IOCTL_EXT_SYNC_SRC,		(void*) _ext_sync_src			},
	{ AIO168_IOCTL_IRQ0_SEL,			(void*) _irq0_sel				},
	{ AIO168_IOCTL_IRQ1_SEL,			(void*) _irq1_sel				},
	{ AIO168_IOCTL_IRQ2_SEL,			(void*) _irq2_sel				},
	{ AIO168_IOCTL_RAG_ENABLE,			(void*) _rag_enable				},
	{ AIO168_IOCTL_RAG_NRATE,			(void*) _rag_nrate				},
	{ AIO168_IOCTL_RANGE,				(void*) _range					},
	{ AIO168_IOCTL_RBG_CLK_SRC,			(void*) _rbg_clk_src			},
	{ AIO168_IOCTL_RBG_ENABLE,			(void*) _rbg_enable				},
	{ AIO168_IOCTL_RBG_NRATE,			(void*) _rbg_nrate				},
	{ AIO168_IOCTL_RX_IO_ABORT,			(void*) _rx_io_abort			},
	{ AIO168_IOCTL_RX_IO_MODE,			(void*) _rx_io_mode				},
	{ AIO168_IOCTL_RX_IO_TIMEOUT,		(void*) _rx_io_timeout			},
	{ AIO168_IOCTL_TX_IO_ABORT,			(void*) _tx_io_abort			},
	{ AIO168_IOCTL_TX_IO_MODE,			(void*) _tx_io_mode				},
	{ AIO168_IOCTL_TX_IO_TIMEOUT,		(void*) _tx_io_timeout			},
	{ AIO168_IOCTL_WAIT_CANCEL,			(void*) gsc_wait_cancel_ioctl	},
	{ AIO168_IOCTL_WAIT_EVENT,			(void*) gsc_wait_event_ioctl	},
	{ AIO168_IOCTL_WAIT_STATUS,			(void*) gsc_wait_status_ioctl	},
	{ -1, NULL }
};


